@@ -268,7 +268,7 @@ pci_device_map_range(struct pci_device *dev, pciaddr_t base,
 
 
     for (region = 0; region < 6; region++) {
-        const struct pci_mem_region const* r = &dev->regions[region];
+        const struct pci_mem_region * const r = &dev->regions[region];
 
         if (r->size != 0) {
             if ((r->base_addr <= base) && ((r->base_addr + r->size) > base)) {
