@@ -568,7 +568,7 @@ pci_system_openbsd_create(void)
 
 	for (domain = 0; domain < sizeof(pcifd) / sizeof(pcifd[0]); domain++) {
 		snprintf(path, sizeof(path), "/dev/pci%d", domain);
-	        pcifd[domain] = open(path, O_RDWR);
+	        pcifd[domain] = open(path, O_RDWR | O_CLOEXEC);
 		if (pcifd[domain] == -1)
 			break;
 		ndomains++;
@@ -733,6 +733,7 @@ int
 pci_device_vgaarb_set_target(struct pci_device *dev)
 {
 	pci_sys->vga_target = dev;
+	return (0);
 }
 
 int
